// DLList.h - Double linked list

/* Author:  Marty Kraimer Date:    26MAR1998 */

/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. MPF is distributed
* subject to a Software License Agreement found in file LICENSE that
* is included with this distribution.
*************************************************************************/
/* This is not a general purpose and therefore "safe"
   implementation of a double link list class. It is not safe for the
   following reasons:

   1) User code directly access obj and casts it to the actual object type.
   2) Many routines will fail if passed a null DLNODE argumment.

   The routines are designed to be fast and easy to use.

   DLList manages a double linked list of DLNode object.
   A DLNode has one public member (void *obj), which MUST be managed by
   use code. The ONLY way to allocate a DLNode is to call DLList::allocNode().
   The ONLY ways to free a DLNode is to call empty or freeNode or delete the
   DLList. The user must cast obj to/from the actual object type for the list.


   typical uses:

   DLList *pmylist = new DLList;
   MyClass *pmyobj;
   ...
   DLNode newnode = DLList::allocNode();
   newnode->obj = (void *)pmyobj;
   pmylist->addEnd(newnode);
   ...
   DLNode *node = pmylist->getFirst();
   while(node) {
       pmyobj = (MyClass *)node->obj;
       ...
       node = pmylist->getNext(node);
   }

    ...
    pmylist->freeNode(node); //after this DON'T use node

*/

#ifndef DLListH
#define DLListH

#include <semLib.h>

class DLNode {
public:
    void   *obj;
private:
    friend class DLList;
    DLNode *next;
    DLNode *prev;
    // disallow default constructor, destructor, copy constructor
    DLNode();
    ~DLNode();
    DLNode(DLNode &);
};

class DLList {
public:
    DLList();
    ~DLList();
    int    getCount();
    DLNode *getFirst();
    DLNode *getLast();
    DLNode *getNext(DLNode *node);
    DLNode *getPrev(DLNode *node);
    void   addEnd(DLNode *node);
    void   addBegin(DLNode *node);
    void   addAfter(DLNode *node,DLNode *newnode);
    void   addBefore(DLNode *node,DLNode *newnode);
    void   empty();
    void   lock();
    void   unlock();
    // The only way to allocate a node is via DLList::allocNode()
    static DLNode *allocNode();
    // freeNode removes node from list and adds node to freelist of nodes
    void   freeNode(DLNode *node);
private:
    DLNode *first;
    DLNode *last;
    int    count;
    SEM_ID sem;
};

inline int     DLList::getCount() {return count;}
inline DLNode *DLList::getFirst() {return first;}
inline DLNode *DLList::getLast() {return last;}
inline DLNode *DLList::getNext(DLNode *node) {return node->next;}
inline DLNode *DLList::getPrev(DLNode *node) {return node->prev;}
inline void    DLList::lock() {semTake(sem,WAIT_FOREVER);}
inline void    DLList::unlock() {semGive(sem);}

#endif //DLListH
